#include <config.h>
! generate a list of three-dimensional paths
!   (M.Hayakawa, November 28, 2010)

module path_generate_class
  use constants_module
!--- module "param_size" was defined in HMCLIB/chrolog_class.F
!  use param_size
  use lattice_class
  use error_class
  use bresenham_module
  implicit none
  private

  public :: disp_set, disp_set_array
  public :: new
  public :: delete
  public :: find
  public :: run
  public :: get_num
  public :: get_disp

  type disp_set
    ! R2 = Lmax^2 + Lmid^2 + Lmin^2
    integer :: R2
    ! List of disp objects with a given R^2
    ! ( There are more than one (Lmax, Lmid, Lmin) such that
    !   Lmax >= Lmid >= Lmin, Lmax^2 + Lmid^2 + Lmin^2 )
    type(disp), pointer :: dp_p => NULL()
  end type disp_set

  type disp_set_array
    ! R(min)^2
    integer :: Rmin2
    ! R(max)^2
    integer :: Rmax2
    ! array of disp_set objects with a specified R^2
    type(disp_set), allocatable :: ds(:)
  end type disp_set_array

  interface new
    module procedure new_disp_set
    module procedure new_disp_set_array
  end interface

  interface delete
    module procedure delete_disp_set
    module procedure delete_disp_set_array
  end interface

  interface find
    module procedure find_disps
  end interface

  interface run
    module procedure run_disp_set_array
  end interface

  interface get_num
    module procedure get_num_of_disps
    module procedure get_num_disp_array
  end interface

  interface get_disp
    module procedure get_disp_pointer_disp_array
  end interface

contains
!---------- IMPLEMENTATION -------------

subroutine new_disp_set( this, R2 )
!  initialize disp_set object (set R2)
!
!  inout:
!    this : disp_set object
!
!  input:
!    R2   : Lmax^2 + Lmid^2 + Lmin^2
!
  implicit none
  type(disp_set), intent(inout) :: this
  integer, intent(in) :: R2

  this%R2 = R2

  if ( .not. associated( this%dp_p ) ) then
    allocate( this%dp_p )
  endif

  return
end subroutine new_disp_set


subroutine new_disp_set_array( this, Rmin2, Rmax2 )
!  initialize disp_set_array object
!     set Rmin2, Rmax2
!     and allocate memory fo array
!
!  inout:
!    this : disp_set object
!
!  input:
!    Rmin2 : minimal value of R of your interest 
!    Rmax2 : maximal value of R of your interest
!           ( required to be less than
!               (NTX/2)^2 + (NTY/2)^2 + (NTZ/2)^2 )
!
  implicit none
  type(disp_set_array), intent(inout) :: this
  integer, intent(in) :: Rmin2, Rmax2
!--------------------------------------
  integer :: k

  if ( Rmin2 <= Rmax2 ) then
    this%Rmin2 = Rmin2
    this%Rmax2 = Rmax2
  else
    this%Rmin2 = Rmax2
    this%Rmax2 = Rmin2
  endif

  k = (NTX/2)**2 + (NTY/2)**2 + (NTZ/2)**2
  if ( this%Rmax2 > k ) then
    this%Rmax2 = k
  endif

  if ( allocated( this%ds ) ) then
    call delete( this )
    deallocate( this%ds )
  endif

  allocate( this%ds( this%Rmin2 : this%Rmax2 ) )

  return
end subroutine new_disp_set_array


subroutine delete_disp_set( this )
! deallocate disp member
!
! inout :
!   this : disp_set object
!
  implicit none
  type(disp_set), intent(inout), target :: this

  call delete( this%dp_p )
  if ( associated( this%dp_p ) ) then
    deallocate( this%dp_p )
    NULLIFY( this%dp_p )
  endif

  return
end subroutine delete_disp_set


subroutine delete_disp_set_array( this )
! deallocate array of disp sets
!
! inout :
!   this : disp_set array object
!
  implicit none
  type(disp_set_array), intent(inout) :: this
!---------------------------
  integer :: j

  if ( allocated( this%ds ) ) then
    do j = this%Rmin2, this%Rmax2
      call delete( this%ds(j) )
    enddo
    deallocate( this%ds )
  endif

  return
end subroutine delete_disp_set_array
 

subroutine find_disps( this, R2 )
! find our all sptial displacements
! by Bresenham algorithm and parmutaion of coodinates
! such that Lmax^2 + Lmid^2 + Lmin^2 = R^2
!
!  inout :
!   this : disp_set object
!
!  input :
!    R2 : R^2
! 
  implicit none
  type(disp_set), intent(inout), target :: this
  integer :: R2
!------------------------------
  type(disp), pointer :: dp_p => NULL(), last_p => NULL()
  integer :: Lmax, Lmid, Lmin
  integer :: Lmax_max, Lmax_min, Lmid_min
  integer :: itmp
  integer :: flag

  call new( this, R2 )

  Lmax_max = AINT( SQRT( REAL( R2, kind=DP) ) )
  Lmax_min = AINT( SQRT( REAL( R2, kind=DP ) /3.0_DP ), kind=DP )

  flag = 0

  do Lmax = Lmax_max, Lmax_min, -1
    Lmid_min = AINT( SQRT( REAL( R2 - Lmax**2, kind=DP ) *0.5_DP ), kind=DP ) 
    do Lmid = Lmax, Lmid_min, -1
      itmp = Lmax**2 + Lmid**2
      do Lmin = Lmid, 0, -1

        if ( ( itmp + Lmin**2 ) .ne. R2 ) then
          cycle
        endif

!!$        ! For debug
!!$        write(*,'("hit (", I5, "):", 3I4)') R2, Lmax, Lmid, Lmin

        if ( flag == 0 ) then
          call new( this%dp_p, Lmax + Lmid + Lmin )
          dp_p => this%dp_p
          flag = 1
        else
          allocate( dp_p%next )
          call new( dp_p%next, Lmax + Lmid + Lmin )
          dp_p => dp_p%next
        endif

        if ( Lmid == 0 ) then
          call run( dp_p, Lmax, last_p )
        else if ( Lmin == 0 ) then
          call run( dp_p, Lmax, Lmid, last_p )
        else
          call run( dp_p, Lmax, Lmid, Lmin, last_p )
        endif

        dp_p => last_p

      enddo ! Lmin
    enddo ! Lmid
  enddo ! Lmax

  return
end subroutine find_disps


function get_num_of_disps( this ) result( num )
! function counting the number of disps in the list
!   of a given disp_set object
!
!  input :
!   this : disp_set object
!
!  output:
!    num : number of disps in the list of disp_set
!
  implicit none
  type(disp_set), intent(in), target :: this
  integer :: num
!-----------------------
  type(disp), pointer :: dp_p

  num = 0
  dp_p => this%dp_p
  do while ( allocated( dp_p%dir ) )
    num = num + 1
    if ( .not. associated(dp_p%next) ) then
      exit
    endif
    dp_p => dp_p%next 
  enddo

  return
end function get_num_of_disps


function get_num_disp_array( this, R2 ) result( num )
! function counting the number of disps in the list
!   of a specified disp_set object
!
!  input :
!   this : disp_set array object
!   R2   : R^2 of your interest
!
!  output:
!    num : number of disps in the list of disp_set of a specified R2
!
  implicit none
  type(disp_set_array), intent(in), target :: this
  integer :: R2
  integer :: num

  num = get_num( this%ds(R2) )

  return
end function get_num_disp_array


subroutine run_disp_set_array( this )
!
! inout :
!   this : disp_set array object
!
  implicit none
  type(disp_set_array), intent(inout) :: this
!---------------------------
  integer :: R2

  if ( .not. allocated( this%ds ) ) then
    call error_stop( "disp_set object has not been initialized!" )
  endif

  do R2 = this%Rmin2, this%Rmax2
    call find( this%ds(R2), R2 )
  enddo

  return
end subroutine run_disp_set_array


function get_disp_pointer_disp_array( this, R2 ) result( dp_p )
! get the pointer of the disp list member of the disp_set with a specified R2
!
!  input :
!   this : disp_set array object
!   R2   : R^2 of your interest
!
!  output:
!   dp_p : pointer of disp member of disp_set with a specified R2
!
  implicit none
  type(disp_set_array), intent(in), target :: this
  integer :: R2
  type(disp), pointer :: dp_p

  dp_p => this%ds(R2)%dp_p

  return
end function get_disp_pointer_disp_array

end module path_generate_class
