module su3exp_class
!
!= su3 matrix exponential and its derivative
!
! Two versions are provided.
!
! 1) Taylor expansion with Cayley-Hamiton theorem
!
! 2) Explicit form without singularity [C.Morningstar & M.Peardon, arXiv:hpe-lat/03110108]
!
  use constants_module
  use lattice_class
  implicit none
  private

  public :: reunit

  public :: get_expm_coefs_taylor_ch
  public :: get_expm_coefs_explicit

  public :: get_expm_taylor_ch
  public :: get_expm_explicit

  !
  ! Taylor expansion order 
  !
  integer, parameter :: EXP_TAYLOR_ORDER = 100

contains

subroutine get_expm_coefs_explicit(A,f,d0f,d1f)
!
! Compute coefficients for matrix exponential/derivative of matrix exponential
! with explicit analytic formula.
!
! exp(i A) = A^2 f(3) + A f(2) + I f(1)
!
!   A : su(3) Hermitian-traceless matrix
!   f : coefficients for exp(i*A)
! d1f : coefficients for dexp(i*A)/dt, d0f = df/dc0
! d0f : coefficients for dexp(i*A)/dt, d1f = df/dc1
!
! c0 = Tr[A^3]/3 = det[A]
! c1 = Tr[A^2]/2
!
  complex(DP),           intent(in)    :: A(COL,COL)
  complex(DP),           intent(inout) ::     f(COL)
  complex(DP), optional, intent(inout) ::   d0f(COL)
  complex(DP), optional, intent(inout) ::   d1f(COL)

  real(DP)    :: c0,c1
  real(DP)    :: rp(COL), sqzp(COL), sqrp(COL)
  complex(DP) :: zp(COL)
  real(DP)    :: rq(COL)
  complex(DP) :: zq(COL),h(COL),ctmp0,ctmp1
  complex(DP) :: r1(COL),r2(COL)
  real(DP)    :: zeta0,zeta1,w2,u2
  real(DP)    :: th,u,w,ev(COL),c0max,rtmp,cosw
  real(DP)    :: rnorm
  logical     :: is_singular
  logical     :: is_inverted
  logical     :: do_derivative

  do_derivative = (present(d0f).and.present(d1f))

  !
  ! Hermitian matrix parametrization
  !
  !  rx(1)    zx(3)^*  zx(2)
  !  zx(3)    rx(2)    zx(1)^*
  !  zx(2)^*  zx(1)    rx(3)
  !

  !
  ! get A
  !
  rp(1) = A(1,1)
  rp(2) = A(2,2)
  rp(3) = -(rp(1)+rp(2))
  zp(1) = A(3,2)
  zp(2) = A(1,3)
  zp(3) = A(2,1)

  sqrp(1) = rp(1)**2
  sqrp(2) = rp(2)**2
  sqrp(3) = rp(3)**2
  sqzp(1) = REAL(zp(1))**2 + AIMAG(zp(1))**2
  sqzp(2) = REAL(zp(2))**2 + AIMAG(zp(2))**2
  sqzp(3) = REAL(zp(3))**2 + AIMAG(zp(3))**2

  rnorm = sqrp(1)+sqrp(2)+sqrp(3) &
 &      + sqzp(1)+sqzp(2)+sqzp(3)
  rnorm = SQRT(rnorm)

  if ( rnorm > EPSILON(1.0_DP)*10 ) then

  !
  ! Squared P
  !
  ! Q = A^2
  !
  rq(1) = sqrp(1) + sqzp(3) + sqzp(2)
  rq(2) = sqzp(3) + sqrp(2) + sqzp(1)
  rq(3) = sqzp(2) + sqzp(1) + sqrp(3)
  zq(1) = zp(1)*(rp(2)+rp(3)) + CONJG(zp(2)*zp(3))
  zq(2) = zp(2)*(rp(3)+rp(1)) + CONJG(zp(3)*zp(1))
  zq(3) = zp(3)*(rp(1)+rp(2)) + CONJG(zp(1)*zp(2))

  !
  ! c1 = Tr[A^2]/2
  !
  c1 = (rq(1) + rq(2) + rq(3))/2.0_DP

  !
  ! c0 = Tr[A^3]/3 = det[A]
  !
  sqrp(1) = sqrp(1)*2.0_DP/3.0_DP
  sqrp(2) = sqrp(2)*2.0_DP/3.0_DP
  sqrp(3) = sqrp(3)*2.0_DP/3.0_DP
  c0 = rp(1)*(rq(1)-sqrp(1))  &
 &   + rp(2)*(rq(2)-sqrp(2))  &
 &   + rp(3)*(rq(3)-sqrp(3))  &
 &   + REAL(zp(1)*zp(2)*zp(3))*2.0_DP

  !----------------------------------------
  ! eigenvalues
  !----------------------------------------

  is_inverted = .FALSE.
  is_singular = .FALSE.

  c0max = 2*SQRT(c1/3.0_DP)**3

  rtmp = c0/c0max
  if (ABS(rtmp) > 1.0_DP) then
    rtmp = SIGN(1.0_DP,rtmp)
  endif
  th = ACOS(rtmp)
  u = SQRT(c1/3.0_DP)*COS(th/3.0_DP)
  w =        SQRT(c1)*SIN(th/3.0_DP)

  if ( ABS(3*u-w) < 0.01_DP ) then

    !
    ! invert c0
    !

    is_inverted = .TRUE.

    c0 = -c0
    rtmp = c0/c0max
    th = ACOS(rtmp)
    u = SQRT(c1/3.0_DP)*COS(th/3.0_DP)
    w =        SQRT(c1)*SIN(th/3.0_DP)
  endif

  u2 = u*u
  w2 = w*w
  ev(1) = 2*u
  ev(2) = -u+w
  ev(3) = -u-w

  !---------------------------
  ! coefficients
  !---------------------------
  if (do_derivative) then
    if ( ABS(w) > 0.05_DP ) then
      zeta0 = SIN(w)/w
      zeta1 = (w*COS(w)-SIN(w))/(w2*w)
    else
      is_singular = .TRUE.
      zeta0 =  1.0_DP-(1.0_DP-(1.0_DP-w2/42.0_DP)*w2/20.0_DP)*w2/6.0_DP
      zeta1 =-(1.0_DP-(1.0_DP-(1.0_DP-w2/54.0_DP)*w2/28.0_DP)*w2/10.0_DP)/3.0_DP
    endif
  else
    if ( ABS(w) > 0.05_DP ) then
      zeta0 = SIN(w)/w
    else
      is_singular = .TRUE.
      zeta0 =  1.0_DP-(1.0_DP-(1.0_DP-w2/42.0_DP)*w2/20.0_DP)*w2/6.0_DP
    endif
  endif

  !-----------------------------
  ! f=(f0,f1,f2)
  !-----------------------------
  ctmp0 = EXP(ZI*(ev(1)))
  ctmp1 = EXP(ZI*(-u))
  cosw  = COS(w)
  h(1) = ((u2-w2)*ctmp0) + (ctmp1*((8*u*cosw + 2*ZI*((3*u2+w2)*zeta0))*u))
  h(2) = (  (2*u)*ctmp0) - (ctmp1* (2*u*cosw -   ZI*((3*u2-w2)*zeta0))   )
  h(3) = (        ctmp0) - (ctmp1* (    cosw +   ZI*( 3*u     *zeta0))   )
  rtmp = 9*u2-w2
  f(1) = h(1)/rtmp
  f(2) = h(2)/rtmp
  f(3) = h(3)/rtmp

  if (do_derivative) then
    !-----------------------------
    ! d0f = df/dc0
    !-----------------------------
    r1(1) = 2*(u     +ZI*(u2-w2))*ctmp0 + ctmp1*( 8*u*(2.0_DP-ZI*u)*cosw + 2*ZI*(9*u2+w2-ZI*u*(3*u2+w2))*zeta0)
    r1(2) = 2*(1.0_DP+ZI*(  2*u))*ctmp0 + ctmp1*(-2  *(1.0_DP-ZI*u)*cosw +   ZI*(6*u    +ZI  *(w2-3*u2))*zeta0)
    r1(3) = 2*        ZI         *ctmp0 + ctmp1*                ZI*(cosw - 3   *(1.0_DP -ZI*u          )*zeta0)

    r2(1) = -2*ctmp0 + 2*ZI*ctmp1*(cosw+(1.0_DP+4*ZI*u)*zeta0 + 3*u2*zeta1)*u
    r2(2) =          -   ZI*ctmp1*(cosw+(1.0_DP+2*ZI*u)*zeta0 - 3*u2*zeta1)
    r2(3) =          -   ZI*ctmp1*(               ZI   *zeta0 + 3*u *zeta1)

    rtmp = 2*rtmp**2
    d0f(1) = (r1(1) - 3*u*r2(1)-24*u*f(1))/rtmp
    d0f(2) = (r1(2) - 3*u*r2(2)-24*u*f(2))/rtmp
    d0f(3) = (r1(3) - 3*u*r2(3)-24*u*f(3))/rtmp

    !-----------------------------
    ! d1f = df/dc1 = H*df/dc0
    !-----------------------------
    d1f(1) = c0*d0f(3)
    d1f(2) = c1*d0f(3)+d0f(1)
    d1f(3) =           d0f(2)
  endif

  if (is_inverted) then
    f(1) =  CONJG(f(1))
    f(2) = -CONJG(f(2))
    f(3) =  CONJG(f(3))

    if (do_derivative) then
      d0f(1) =-CONJG(d0f(1))
      d0f(2) = CONJG(d0f(2))
      d0f(3) =-CONJG(d0f(3))
      d1f(1) = CONJG(d1f(1))
      d1f(2) =-CONJG(d1f(2))
      d1f(3) = CONJG(d1f(3))
    endif
  endif

  else ! else of if (rnrom ...)

    f(:) = Z0
    f(1) = Z1

    if (do_derivative) then
      d0f(:) = Z0
      d1f(:) = Z0
    endif

  endif ! end of if (rnrom ...)

  return
end subroutine

subroutine get_expm_explicit(A,expA,dA,dexpA)
!
!     A : 
!  expA :   exp(i*A)
!    dA :         dA/dt
! dexpA :  dexp(i*A)/dt
!
  implicit none
  complex(DP), intent(in)    ::     A(COL,COL)
  complex(DP), intent(inout) ::  expA(COL,COL)
  complex(DP), optional, intent(in)    ::    dA(COL,COL)
  complex(DP), optional, intent(inout) :: dexpA(COL,COL)
  
  real(DP)    :: dc0,dc1
  real(DP)    :: rp(COL), sqzp(COL), sqrp(COL)
  complex(DP) :: zp(COL)
  real(DP)    :: rq(COL)
  complex(DP) :: zq(COL),f(COL),h(COL),ctmp0,ctmp1
  complex(DP) :: d0f(COL),d1f(COL),df(COL)


  !
  ! Hermitian matrix parametrization
  !
  !  rx(1)    zx(3)^*  zx(2)
  !  zx(3)    rx(2)    zx(1)^*
  !  zx(2)^*  zx(1)    rx(3)
  !

  !
  ! get A
  !
  rp(1) = A(1,1)
  rp(2) = A(2,2)
  rp(3) = -(rp(1)+rp(2))
  zp(1) = A(3,2)
  zp(2) = A(1,3)
  zp(3) = A(2,1)

  sqrp(1) = rp(1)**2
  sqrp(2) = rp(2)**2
  sqrp(3) = rp(3)**2
  sqzp(1) = REAL(zp(1))**2 + AIMAG(zp(1))**2
  sqzp(2) = REAL(zp(2))**2 + AIMAG(zp(2))**2
  sqzp(3) = REAL(zp(3))**2 + AIMAG(zp(3))**2

  !
  ! Squared P
  !
  ! Q = A^2
  !
  rq(1) = sqrp(1) + sqzp(3) + sqzp(2)
  rq(2) = sqzp(3) + sqrp(2) + sqzp(1)
  rq(3) = sqzp(2) + sqzp(1) + sqrp(3)
  zq(1) = zp(1)*(rp(2)+rp(3)) + CONJG(zp(2)*zp(3))
  zq(2) = zp(2)*(rp(3)+rp(1)) + CONJG(zp(3)*zp(1))
  zq(3) = zp(3)*(rp(1)+rp(2)) + CONJG(zp(1)*zp(2))

  call get_expm_coefs_explicit(A,f,d0f,d1f)

  expA(1,1) = f(3)*      rq(1)  + f(2)*      rp(1) + f(1)
  expA(2,1) = f(3)*      zq(3)  + f(2)*      zp(3)
  expA(3,1) = f(3)*CONJG(zq(2)) + f(2)*CONJG(zp(2))

  expA(1,2) = f(3)*CONJG(zq(3)) + f(2)*CONJG(zp(3))
  expA(2,2) = f(3)*      rq(2)  + f(2)*      rp(2) + f(1)
  expA(3,2) = f(3)*      zq(1)  + f(2)*      zp(1)

  expA(1,3) = f(3)*      zq(2)  + f(2)*      zp(2)
  expA(2,3) = f(3)*CONJG(zq(1)) + f(2)*CONJG(zp(1))
  expA(3,3) = f(3)*      rq(3)  + f(2)*      rp(3) + f(1)

  if (present(dA).and.present(dexpA)) then
    !
    ! dc1 = Tr[dA.A]
    !
    dc1 =   REAL(dA(1,1),kind=DP)*rp(1) &
 &         +REAL(dA(2,2),kind=DP)*rp(2) &
 &         +REAL(dA(3,3),kind=DP)*rp(3) &
 &      +2*(REAL(dA(1,2)*zp(3),kind=DP) &
 &         +REAL(dA(2,3)*zp(1),kind=DP) &
 &         +REAL(dA(3,1)*zp(2),kind=DP))

    !
    ! dc0 = Tr[dA.A^2]
    !
    dc0 =   REAL(dA(1,1),kind=DP)*rq(1) &
 &         +REAL(dA(2,2),kind=DP)*rq(2) &
 &         +REAL(dA(3,3),kind=DP)*rq(3) &
 &      +2*(REAL(dA(1,2)*zq(3),kind=DP) &
 &         +REAL(dA(2,3)*zq(1),kind=DP) &
 &         +REAL(dA(3,1)*zq(2),kind=DP))

    df(:) = d0f(:)*dc0 + d1f(:)*dc1
    dexpA(1,1) = df(3)*      rq(1)  + df(2)*      rp(1) + df(1)
    dexpA(2,1) = df(3)*      zq(3)  + df(2)*      zp(3)
    dexpA(3,1) = df(3)*CONJG(zq(2)) + df(2)*CONJG(zp(2))

    dexpA(1,2) = df(3)*CONJG(zq(3)) + df(2)*CONJG(zp(3))
    dexpA(2,2) = df(3)*      rq(2)  + df(2)*      rp(2) + df(1)
    dexpA(3,2) = df(3)*      zq(1)  + df(2)*      zp(1)

    dexpA(1,3) = df(3)*      zq(2)  + df(2)*      zp(2)
    dexpA(2,3) = df(3)*CONJG(zq(1)) + df(2)*CONJG(zp(1))
    dexpA(3,3) = df(3)*      rq(3)  + df(2)*      rp(3) + df(1)

    dexpA(:,:) = dexpA(:,:) + f(3)*(MATMUL(dA,A)+MATMUL(A,dA))+f(2)*dA(:,:)
  endif

  return
end subroutine

subroutine get_expm_coefs_taylor_ch(A,f,d0f,d1f)
!
! Compute coefficients for matrix exponential/derivative of matrix exponential
! with Taylor expansion (augmented matrix method).
!
! exp(i A) = A^2 f(3) + A f(2) + I f(1)
!
!   A : su(3) Hermitian-traceless matrix
!   f : coefficients for exp(i*A)
! d0f : coefficients for dexp(i*A)/dt, d0f = df/dc0
! d1f : coefficients for dexp(i*A)/dt, d1f = df/dc1
!
! c0 = Tr[A^3]/3 = det[A]
! c1 = Tr[A^2]/2
!
  implicit none
  complex(DP),           intent(in)    :: A(COL,COL)
  complex(DP),           intent(inout) ::     f(COL)
  complex(DP), optional, intent(inout) ::   d0f(COL)
  complex(DP), optional, intent(inout) ::   d1f(COL)
  integer :: ic,jc,iexp
  complex(DP) :: aa(COL),bb(COL),zcoef
  real(DP) :: c0,c1,dc0,dc1
  real(DP)    :: rp(COL), sqzp(COL), sqrp(COL)
  complex(DP) :: zp(COL)
  real(DP)    :: rq(COL)
  complex(DP) :: zq(COL)

  !
  ! Hermitian matrix parametrization
  !
  !  rx(1)    zx(3)^*  zx(2)
  !  zx(3)    rx(2)    zx(1)^*
  !  zx(2)^*  zx(1)    rx(3)
  !

  !
  ! P <= A
  !
  rp(1) = A(1,1)
  rp(2) = A(2,2)
  rp(3) = -(rp(1)+rp(2))
  zp(1) = A(3,2)
  zp(2) = A(1,3)
  zp(3) = A(2,1)

  sqrp(1) = rp(1)**2
  sqrp(2) = rp(2)**2
  sqrp(3) = rp(3)**2
  sqzp(1) = REAL(zp(1))**2 + AIMAG(zp(1))**2
  sqzp(2) = REAL(zp(2))**2 + AIMAG(zp(2))**2
  sqzp(3) = REAL(zp(3))**2 + AIMAG(zp(3))**2

  !
  ! Squared P
  !
  ! Q = A^2
  !
  rq(1) = sqrp(1) + sqzp(3) + sqzp(2)
  rq(2) = sqzp(3) + sqrp(2) + sqzp(1)
  rq(3) = sqzp(2) + sqzp(1) + sqrp(3)
  zq(1) = zp(1)*(rp(2)+rp(3)) + CONJG(zp(2)*zp(3))
  zq(2) = zp(2)*(rp(3)+rp(1)) + CONJG(zp(3)*zp(1))
  zq(3) = zp(3)*(rp(1)+rp(2)) + CONJG(zp(1)*zp(2))

  !
  ! c1 = Tr[A^2]/2
  !
  c1 = (rq(1) + rq(2) + rq(3))/2.0_DP

  !
  ! c0 = Tr[A^3]/3
  !
  sqrp(1) = sqrp(1)*2.0_DP/3.0_DP
  sqrp(2) = sqrp(2)*2.0_DP/3.0_DP
  sqrp(3) = sqrp(3)*2.0_DP/3.0_DP
  c0 = rp(1)*(rq(1)-sqrp(1))  &
 &   + rp(2)*(rq(2)-sqrp(2))  &
 &   + rp(3)*(rq(3)-sqrp(3))  &
 &   + REAL(zp(1)*zp(2)*zp(3))*2.0_DP

  if (present(d0f).and.present(d1f)) then
      f(1) = Z1
      f(2) = Z0
      f(3) = Z0
    d0f(1) = Z0
    d0f(2) = Z0
    d0f(3) = Z0
    do iexp = EXP_TAYLOR_ORDER,1,-1
      zcoef = ZI/iexp
      !--------------------------------
      !      | 0  0  c0 |
      ! H  = | 1  0  c1 |
      !      | 0  1   0 |
      !
      !      | 0  0   1 |
      ! H' = | 0  0   0 | = dH/dc0
      !      | 0  0   0 |
      !
      !
      !   a = H f
      !   b = H d0f + H' f
      !--------------------------------
      aa(1) = c0* f(3)
      aa(2) = c1* f(3)+ f(1)
      aa(3) =           f(2)
      bb(1) = c0*d0f(3)       +f(3)
      bb(2) = c1*d0f(3)+d0f(1)
      bb(3) =           d0f(2)

      !--------------------------------
      !   f = f0 + i*H/j f
      !     = f0 + i*a/j
      !
      ! d0f = i H/j d0f + i H'/j f
      !     = i*b/j
      !--------------------------------
       f(1)  = aa(1)*zcoef + Z1
       f(2)  = aa(2)*zcoef
       f(3)  = aa(3)*zcoef
      d0f(1) = bb(1)*zcoef
      d0f(2) = bb(2)*zcoef
      d0f(3) = bb(3)*zcoef
    enddo

    !------------------------
    !  d1f = H d0f
    !------------------------
    d1f(1) = c0*d0f(3)
    d1f(2) = c1*d0f(3) + d0f(1)
    d1f(3) =           + d0f(2)
  else

    f(1) = Z1
    f(2) = Z0
    f(3) = Z0
    do iexp = EXP_TAYLOR_ORDER,1,-1
      zcoef = ZI/iexp
      !--------------------------------
      !      | 0  0  c0 |
      ! H  = | 1  0  c1 |
      !      | 0  1   0 |
      !
      !   a = H f
      !--------------------------------
      aa(1) = c0* f(3)
      aa(2) = c1* f(3)+ f(1)
      aa(3) =           f(2)

      !--------------------------------
      !   f = f0 + i*H/j f
      !     = f0 + i*a/j
      !--------------------------------
      f(1)  = aa(1)*zcoef + Z1
      f(2)  = aa(2)*zcoef
      f(3)  = aa(3)*zcoef
    enddo

  endif

  return
end subroutine

subroutine get_expm_taylor_ch(A,expA,dA,dexpA)
!
! compute SU(3) matrix via exponentiating with su(3) (Hermitian traceless) matrix P times i.
!
! expA = exp(i*A)
!
!    A : su(3) matrix.
! expA : SU(3) matrix.
!
  implicit none
  complex(DP), intent(in)    ::     A(COL,COL)
  complex(DP), intent(inout) ::  expA(COL,COL)
  complex(DP), optional, intent(in)    ::    dA(COL,COL)
  complex(DP), optional, intent(inout) :: dexpA(COL,COL)
  integer :: ic,jc
  complex(DP) :: f(COL),df(COL)
  complex(DP) :: d0f(COL),d1f(COL)
  real(DP)    :: dc0,dc1
  real(DP)    :: rp(COL), sqzp(COL), sqrp(COL)
  complex(DP) :: zp(COL)
  real(DP)    :: rq(COL)
  complex(DP) :: zq(COL)

  !
  ! Hermitian matrix parametrization
  !
  !  rx(1)    zx(3)^*  zx(2)
  !  zx(3)    rx(2)    zx(1)^*
  !  zx(2)^*  zx(1)    rx(3)
  !

  !
  ! P <= A
  !
  rp(1) = A(1,1)
  rp(2) = A(2,2)
  rp(3) = -(rp(1)+rp(2))
  zp(1) = A(3,2)
  zp(2) = A(1,3)
  zp(3) = A(2,1)

  sqrp(1) = rp(1)**2
  sqrp(2) = rp(2)**2
  sqrp(3) = rp(3)**2
  sqzp(1) = REAL(zp(1))**2 + AIMAG(zp(1))**2
  sqzp(2) = REAL(zp(2))**2 + AIMAG(zp(2))**2
  sqzp(3) = REAL(zp(3))**2 + AIMAG(zp(3))**2

  !
  ! Squared P
  !
  ! Q = A^2
  !
  rq(1) = sqrp(1) + sqzp(3) + sqzp(2)
  rq(2) = sqzp(3) + sqrp(2) + sqzp(1)
  rq(3) = sqzp(2) + sqzp(1) + sqrp(3)
  zq(1) = zp(1)*(rp(2)+rp(3)) + CONJG(zp(2)*zp(3))
  zq(2) = zp(2)*(rp(3)+rp(1)) + CONJG(zp(3)*zp(1))
  zq(3) = zp(3)*(rp(1)+rp(2)) + CONJG(zp(1)*zp(2))

  call get_expm_coefs_taylor_ch(A,f,d0f,d1f)


  expA(1,1) = f(3)*      rq(1)  + f(2)*      rp(1) + f(1)
  expA(2,1) = f(3)*      zq(3)  + f(2)*      zp(3)
  expA(3,1) = f(3)*CONJG(zq(2)) + f(2)*CONJG(zp(2))

  expA(1,2) = f(3)*CONJG(zq(3)) + f(2)*CONJG(zp(3))
  expA(2,2) = f(3)*      rq(2)  + f(2)*      rp(2) + f(1)
  expA(3,2) = f(3)*      zq(1)  + f(2)*      zp(1)

  expA(1,3) = f(3)*      zq(2)  + f(2)*      zp(2)
  expA(2,3) = f(3)*CONJG(zq(1)) + f(2)*CONJG(zp(1))
  expA(3,3) = f(3)*      rq(3)  + f(2)*      rp(3) + f(1)

  if (present(dA).and.present(dexpA)) then
    !
    ! dc1 = Tr[dA.A]
    !
    dc1 =   REAL(dA(1,1),kind=DP)*rp(1) &
 &         +REAL(dA(2,2),kind=DP)*rp(2) &
 &         +REAL(dA(3,3),kind=DP)*rp(3) &
 &      +2*(REAL(dA(1,2)*zp(3),kind=DP) &
 &         +REAL(dA(2,3)*zp(1),kind=DP) &
 &         +REAL(dA(3,1)*zp(2),kind=DP))

    !
    ! dc0 = Tr[dA.A^2]
    !
    dc0 =   REAL(dA(1,1),kind=DP)*rq(1) &
 &         +REAL(dA(2,2),kind=DP)*rq(2) &
 &         +REAL(dA(3,3),kind=DP)*rq(3) &
 &      +2*(REAL(dA(1,2)*zq(3),kind=DP) &
 &         +REAL(dA(2,3)*zq(1),kind=DP) &
 &         +REAL(dA(3,1)*zq(2),kind=DP))
    !----------------------------
    ! df = dc0*d0f + dc1 * d1f
    !----------------------------
    df(:) = dc0 * d0f(:) + dc1 * d1f(:)

    dexpA(1,1) = df(3)*      rq(1)  + df(2)*      rp(1) + df(1)
    dexpA(2,1) = df(3)*      zq(3)  + df(2)*      zp(3)
    dexpA(3,1) = df(3)*CONJG(zq(2)) + df(2)*CONJG(zp(2))

    dexpA(1,2) = df(3)*CONJG(zq(3)) + df(2)*CONJG(zp(3))
    dexpA(2,2) = df(3)*      rq(2)  + df(2)*      rp(2) + df(1)
    dexpA(3,2) = df(3)*      zq(1)  + df(2)*      zp(1)

    dexpA(1,3) = df(3)*      zq(2)  + df(2)*      zp(2)
    dexpA(2,3) = df(3)*CONJG(zq(1)) + df(2)*CONJG(zp(1))
    dexpA(3,3) = df(3)*      rq(3)  + df(2)*      rp(3) + df(1)

    dexpA(:,:) = dexpA(:,:) + f(3)*(MATMUL(dA,A)+MATMUL(A,dA))+f(2)*dA(:,:)

  endif

  return
end subroutine

subroutine reunit(u)
!
! SU(3) matrix unitarization
!
  implicit none
  complex(DP), intent(inout) :: u(COL,COL)
  complex(DP) :: ctmp
  real(DP) :: rnorm
  rnorm = REAL(u(1,1))**2 + AIMAG(u(1,1))**2 &
 &      + REAL(u(2,1))**2 + AIMAG(u(2,1))**2 &
 &      + REAL(u(3,1))**2 + AIMAG(u(3,1))**2
  rnorm = 1.0_DP/SQRT(rnorm)
  u(1,1) = u(1,1)*rnorm
  u(2,1) = u(2,1)*rnorm
  u(3,1) = u(3,1)*rnorm
  ctmp = CONJG(u(1,1)) * u(1,2)  &
 &     + CONJG(u(2,1)) * u(2,2)  &
 &     + CONJG(u(3,1)) * u(3,2)
  u(1,2) = u(1,2) - u(1,1)*ctmp
  u(2,2) = u(2,2) - u(2,1)*ctmp
  u(3,2) = u(3,2) - u(3,1)*ctmp
  rnorm = REAL(u(1,2))**2 + AIMAG(u(1,2))**2 &
 &      + REAL(u(2,2))**2 + AIMAG(u(2,2))**2 &
 &      + REAL(u(3,2))**2 + AIMAG(u(3,2))**2
  rnorm = 1.0_DP/SQRT(rnorm)
  u(1,2) = u(1,2)*rnorm
  u(2,2) = u(2,2)*rnorm
  u(3,2) = u(3,2)*rnorm
  u(1,3) = CONJG( u(2,1)*u(3,2) - u(3,1)*u(2,2) )
  u(2,3) = CONJG( u(3,1)*u(1,2) - u(1,1)*u(3,2) )
  u(3,3) = CONJG( u(1,1)*u(2,2) - u(2,1)*u(1,2) )
  return
end subroutine

function get_matrix_fnorm(A) result(fnorm)
!
! compute Frobenius norm for su(3) matrix (Hermitian traceless)
!
  implicit none
  complex(DP), intent(in) :: A(COL,COL)
  real(DP) :: fnorm
  integer :: i
  fnorm = (real(A(1,1),kind=DP)**2  &
 &       + real(A(2,2),kind=DP)**2  &
 &       + real(A(1,1),kind=DP)*real(A(2,2),kind=DP)   &
 &       + real(A(2,1),kind=DP)**2 + aimag(A(2,1))**2  &
 &       + real(A(3,1),kind=DP)**2 + aimag(A(3,1))**2  &
 &       + real(A(3,2),kind=DP)**2 + aimag(A(3,2))**2)*2.0_DP
  fnorm = SQRT(fnorm)
  return
end function

function get_scaling_squaring_step(fnorm) result(nstep)
!
! compute number of scaling-squaring step to satisfy
!
! fnorm/(2**nstep) < 1/2
!
  implicit none
  real(DP), intent(in) :: fnorm
  integer :: nstep
  nstep = MAX(CEILING(LOG(2*fnorm)/LOG(2.0_DP)),0)
  return
end function

end module
