#include <config.h>
!#define _DEBUG
module ks_class
!
!= Krylov-Schur Algorithm for Large eigenvalue problem test module
!  2009/3/13 Ken-Ichi Ishikawa
!
!== Version
!  $Id: ks_class.F90,v 1.5 2011/06/08 03:03:24 ishikawa Exp $
!
!== Refs.:
!
!* G. W. Stewart,
! "A Krylov--Schur Algorithm for Large Eigenproblems",
!  SIAM Journal on Matrix Analysis and Applications,
!  Volume 23 (2001) 601-614.  
!
!* G. W. Stewart,
!  "Addendum to "A Krylov--Schur Algorithm for Large Eigenproblems""
!  SIAM Journal on Matrix Analysis and Applications,
!  Volume 24 (2002) 599-601, 
!
!* V. Hernandez, J. E. Roman, A. Tomas, and V. Vidal,
! "Krylov-Schur Methods in SLEPc",
!  Scalable Library for Eigenvalue Problem Computations
!  SLEPc Technical Report STR-7,
!  [http://www.grycap.upv.es/slepc].
!
!============================================================================
  use constants_module
  use comlib
  use lattice_class, only : nodeid
  use error_class
  implicit none
  private
  public :: ks_alg
  public :: get_status
  public :: get_name
  public :: get_current_iteration
  public :: get_num_converged
  public :: new
  public :: delete
  public :: solve
  public :: ESLV_OP_NOP
  public :: ESLV_OP_DO_MATVEC
  public :: ESLV_OP_FAIL
  public :: ESLV_OP_CONVERGED

  public :: ESLV_MODE_GENERAL
  public :: ESLV_MODE_HERMITE

  integer, parameter ::   EV_CONVED=0
  integer, parameter ::   EV_TARGET=1
  integer, parameter ::   EV_UNCONV=2

  integer, parameter ::   ESLV_OP_NOP       = 1
  integer, parameter ::   ESLV_OP_DO_MATVEC = 2
  integer, parameter ::   ESLV_OP_CONVERGED = 3
  integer, parameter ::   ESLV_OP_FAIL      = -9
  integer, parameter ::   ESLV_OP_END_ARNOLDI = 10

  integer, parameter ::   ESLV_MODE_GENERAL = 1
  integer, parameter ::   ESLV_MODE_HERMITE = 2

  logical :: MODE_NORMAL = .true.

  type arnoldi_alg
    complex(DP), pointer :: V(:,:) => NULL()
    complex(DP), pointer :: H(:,:) => NULL()
    complex(DP), pointer :: src_vec(:) => NULL()
    complex(DP), pointer :: dst_vec(:) => NULL()
    integer :: n = 0
    integer :: m = 0
    integer :: k = 0
    integer :: j = 0
    integer :: mode = ESLV_MODE_GENERAL
    integer :: istat = 0
    integer :: inext = 0
    integer :: idummy = 0
  end type

  type ks_alg
    type(arnoldi_alg), private :: arnoldi
    complex(DP), pointer, public :: src_vec(:) => NULL()
    complex(DP), pointer, public :: dst_vec(:) => NULL()
    complex(DP), pointer, public :: eval(:) => NULL()
    complex(DP), pointer, public :: evec(:,:) => NULL()
    complex(DP), pointer, public :: V(:,:) => NULL()
    complex(DP), pointer, private :: r(:)   => NULL()
    complex(DP), pointer, private :: H(:,:) => NULL()
    complex(DP), pointer, private :: T(:,:) => NULL()
    complex(DP), pointer, private :: Z(:,:) => NULL()
    real(DP), pointer, private :: res(:) => NULL()
    integer,  pointer, private :: tag(:) => NULL()
    real(DP), private :: tol = 1.0_DP
    logical, private :: debug = .false.
    integer, private :: n = 0
    integer, private :: m = 0
    integer, private :: k = 0
    integer, private :: nev = 0
    integer, private :: icnv  = 0
    integer, private :: icnv0 = 0
    integer, private :: inext = 0
    integer, private :: istat = 0
    integer, private :: iter = 0
    integer, private :: maxiter=1000
    integer, private :: mode = ESLV_MODE_GENERAL
    character(256), private :: name="Krylov-Schur Algorithm"
  end type

  interface new
    module procedure new_ks
  end interface

  interface delete
    module procedure delete_ks
  end interface

  interface solve
    module procedure solve_ks
  end interface

  interface get_status
    module procedure get_status_ks
  end interface

  interface get_current_iteration
    module procedure get_current_iteration_ks
  end interface

  interface get_name
    module procedure get_name_ks
  end interface

  interface get_num_converged
    module procedure get_num_convd_ks
  end interface

contains

subroutine new_ks(this,n,m,nev,tol,maxiter,debug,mode)
  implicit none
  type(ks_alg), intent(inout) :: this
  integer, intent(in) :: n,m,nev,maxiter
  real(DP), intent(in) :: tol
  logical, optional, intent(in) :: debug
  integer, optional, intent(in) :: mode
  character(len=CHARLEN) :: str
  this%n = n   ! matrix size
  this%m = m   ! Krylov-subspace size
  this%k = 0
  this%nev = nev  ! number of eigen values wanted
  this%tol = tol
  this%iter = 1
  this%maxiter = maxiter
  this%inext = 0
  this%icnv  = 0
  this%icnv0 = 0
  if (m < nev) then
    write(str,'(" Krylov-subspace dimension is smaller than required number of eigen values.")')
    call error_stop(str)
  endif
  if (n < m) then
    write(str,'(" matrix size is smaller than Krylov-subspace dimension.")')
    call error_stop(str)
  endif
  if (present(debug)) then
    this%debug = debug
  else
    this%debug = .false.
  endif
  if (present(mode)) then
    this%mode = mode
  else
    this%mode = ESLV_MODE_GENERAL
  endif

  allocate(this%V(n,m+1),this%r(n))
  allocate(this%H(m+1,m))
  allocate(this%T(m,m))
  allocate(this%Z(m,m))
  allocate(this%eval(m),this%res(m),this%tag(m))
  allocate(this%evec(m,m))

  this%arnoldi%V => NULL()
  this%arnoldi%H => NULL()
  this%arnoldi%mode = this%mode
  this%src_vec => NULL()
  this%dst_vec => NULL()
  this%H(:,:) = Z0
  this%T(:,:) = Z0
  this%Z(:,:) = Z0
  this%V(:,1) = Z1
  this%eval(:) = Z1
  this%evec(:,:) = Z1
  this%tag(:)  = EV_UNCONV
  return
end subroutine

subroutine delete_ks(this)
  type(ks_alg), intent(inout) :: this
  deallocate(this%V,this%r)
  deallocate(this%H)
  deallocate(this%T)
  deallocate(this%Z)
  deallocate(this%eval,this%evec,this%res,this%tag)
  return
end subroutine

function get_status_ks(this) result(status)
  type(ks_alg), intent(in) :: this
  integer :: status
  status = this%istat
  return
end function

function get_name_ks(this) result(name)
  type(ks_alg), intent(in) :: this
  character(256) :: name
  name = this%name
  return
end function

function get_current_iteration_ks(this) result(iter)
  type(ks_alg), intent(in) :: this
  integer :: iter
  iter = this%iter
  return
end function

function get_num_convd_ks(this) result(ncnv)
  type(ks_alg), intent(in) :: this
  integer :: ncnv
  ncnv = this%icnv
  return
end function

subroutine solve_ks(this)
  implicit none
  type(ks_alg), intent(inout) :: this
  complex(DP), allocatable :: Q(:,:)
  complex(DP) :: ctmp
  real(DP) :: rtmp
  integer :: i,j,jj
  integer :: k,n,m,is

  select case(this%inext)
  case (0)
    k = this%k
    n = this%n
    m = this%m

    if (this%debug .and. 0 == nodeid) &
 &    write(*,'("%Krylov-Schur iteration:",I5)')this%iter

    rtmp = norm(this%V(:,k+1))
    rtmp = 1.0_DP/rtmp
!$OMP PARALLEL DO PRIVATE(is)
    do is=1,this%n
      this%V(is,k+1) = this%V(is,k+1)*rtmp
    enddo

    !=========================================================================
    ! Compute Arnoldi factorization,  A \ V_m = V_m+1 H_m
    ! invert method
    !=========================================================================

    call new_arnoldi(this%arnoldi,n,m,k)
    this%inext = this%inext + 1
    this%istat = ESLV_OP_NOP
    this%arnoldi%V => this%V
    this%arnoldi%H => this%H

    return

  case (1)

    call run_arnoldi(this%arnoldi)

    select case(this%arnoldi%istat)
    case (ESLV_OP_NOP)
      this%inext = 1
      this%istat = ESLV_OP_NOP
      return
    case (ESLV_OP_DO_MATVEC)
      this%src_vec => this%arnoldi%src_vec
      this%dst_vec => this%arnoldi%dst_vec
      this%inext = 1
      this%istat = ESLV_OP_DO_MATVEC
      if (this%debug .and. 0 == nodeid) &
 &      write(*,'("Arnoldi iteration.",I4)')this%arnoldi%j
      return
    case (ESLV_OP_END_ARNOLDI)
      this%inext = 2
      this%istat = ESLV_OP_NOP
      if (this%debug .and. 0 == nodeid) &
 &      write(*,'("Arnoldi iteration ENDED.")')
      return
    end select

  case (2)

    m = this%m
    if (this%debug .and. 0 == nodeid) &
 &    write(*,'("%Schur transformation")')
    !=========================================================================
    ! Compute Schur decomposition of H_m = Z_m * T_m * Z_m'
    !
    !  T : upper triangular (sorted)
    !  Z : orthogonal Schur basis
    !=========================================================================
#ifdef _GGGDEBUG
    if (this%icnv >0) then
      this%T(1:m,1:m) = this%H(1:m,1:m)
      this%Z(1:m,1:m) = Z0
      call get_schur(this%icnv,m,this%T,this%eval,this%Z,ESLV_MODE_HERMITE)
      if (0 == nodeid) then
        do j=1,m
        do i=1,m
          write(*,'("HHH",2I3,2E14.6)')i,j,this%T(i,j)
        enddo
        enddo
      endif
    endif
#endif

    this%T(1:m,1:m) = this%H(1:m,1:m)
    this%Z(1:m,1:m) = Z0
    call get_schur(this%icnv,m,this%T,this%eval,this%Z,this%mode)
#ifdef _GGGDEBUG
    if (this%icnv >0) then
    if (0==nodeid) then
      do j=1,m
      do i=1,m
        write(*,'("GGG",2I3,2E14.6)')i,j,this%T(i,j)
      enddo
      enddo
      stop
    endif
    endif
#endif


    !=========================================================================
    ! Krylov-Schur transformation
    !  _    _
    !  H <= H Z
    !=========================================================================
    this%H(1:m,1:m)=this%T(1:m,1:m)
    rtmp = this%H(m+1,m)
    do i=1,m
      this%H(m+1,i) = rtmp*this%Z(m,i)
    enddo

    !=========================================================================
    ! Construct Schur vector, V <= V*Z
    !=========================================================================
    n = this%n
    allocate(Q(n,m))
    do i=this%icnv+1,m
!$OMP PARALLEL DO PRIVATE(jj,j)
      do jj=1,this%n
        Q(jj,i) = this%V(jj,this%icnv+1)*this%Z(this%icnv+1,i)
        do j=this%icnv+2,m
          Q(jj,i) = Q(jj,i) + this%V(jj,j)*this%Z(j,i)
        enddo
      enddo
!$OMP END PARALLEL DO
    enddo
    do i=this%icnv+1,m
!$OMP PARALLEL DO PRIVATE(is)
    do is=1,this%n
      this%V(is,i)=Q(is,i)
    enddo
    enddo
    deallocate(Q)

    this%T(1:m,1:m) = this%H(1:m,1:m)
    this%Z(1:m,1:m) = Z0
    call get_eig(this%icnv,m,this%T,this%eval,this%evec,this%mode)

    this%tag(:) = EV_UNCONV
    do i=1,m
      ctmp = Z0
      do j=1,m
        ctmp = ctmp + conjg(this%H(m+1,j))*this%evec(j,i)
      enddo
      this%res(i) = abs(ctmp)/abs(this%eval(i))
      if (this%res(i) < this%tol) then
        this%tag(i) = EV_CONVED
        if (this%debug .and. 0 == nodeid) &
  &        write(*,'("%E",I4,3E24.15," T")')i,this%eval(i),this%res(i)
      else
        if (this%debug .and. 0 == nodeid) &
  &        write(*,'("%E",I4,3E24.15," F")')i,this%eval(i),this%res(i)
      endif
    enddo

    this%icnv=0
    do i=1,m
      if (this%tag(i) == EV_CONVED) then
        this%icnv=this%icnv+1
      else
        exit
      endif
    enddo
    this%tag(:) = EV_UNCONV
    do i=1,this%icnv
      this%tag(i) = EV_CONVED
      this%H(m+1,i) = Z0
    enddo

    if (this%icnv >= this%nev) then
      this%istat = ESLV_OP_NOP
      this%inext = 4
      return
    endif

    if (this%debug .and. 0 == nodeid) write(*,'("%Purge")')
    !=========================================================================
    ! Purge Schur decomposition
    !=========================================================================
    k= MIN(this%nev+this%icnv,m-1)
    this%k = k
    do i=1,k
      this%H(k+1,i)=this%H(m+1,i)
      !
      ! converged subspace are decoupled.
      !
      if (this%tag(i) == EV_CONVED) this%H(k+1,i) = Z0
      if (this%debug .and. 0 == nodeid) then
        write(*,'("%NEV_ICNV_K",3I4)')this%nev,this%icnv,k    
        write(*,'("%T",I3,2E15.7)')i,this%H(k+1,i)
      endif
    enddo
!$OMP PARALLEL DO PRIVATE(is)
    do is=1,this%n
      this%V(is,k+1) = this%V(is,m+1)
    enddo
    this%H(  1:m+1,k+1:m) = Z0
    this%H(k+2:m+1,1:k+1) = Z0


    this%icnv0 = this%icnv

    this%iter = this%iter + 1
    this%inext = 0
    this%istat = ESLV_OP_NOP

    if (this%iter > this%maxiter) then
      this%inext = 0
      this%istat = ESLV_OP_FAIL
      return
    endif

    return

  case (4)

    call sort_eig(this%icnv,this%m,this%eval,this%n,this%V)

    this%istat = ESLV_OP_CONVERGED
    this%inext = 0
    return
  end select

  return
end subroutine

subroutine new_arnoldi(this,n,m,k)
  implicit none
  type(arnoldi_alg), intent(inout) :: this
  integer, intent(in) :: n,m,k
  this%j = k+1
  this%n = n
  this%m = m
  this%k = k
  return
end subroutine

subroutine run_arnoldi(this)
  implicit none
  type(arnoldi_alg), intent(inout) :: this
  real(DP) :: rtmp
  complex(DP) :: ctmp
  integer :: i,j,is

  select case(this%inext)
  case(0)
    this%istat = ESLV_OP_DO_MATVEC
    this%inext = this%inext+1
    this%src_vec => this%V(:,this%j)
    this%dst_vec => this%V(:,this%j+1)
    return
  case(1)
    j=this%j
    do i=1,j
      this%H(i,j) = prod(this%V(:,i),this%V(:,j+1))
!$OMP PARALLEL DO PRIVATE(is)
      do is=1,this%n
        this%V(is,j+1) = this%V(is,j+1) - this%V(is,i)*this%H(i,j)
      enddo
    enddo
    do i=1,j
      ctmp = prod(this%V(:,i),this%V(:,j+1))
!$OMP PARALLEL DO PRIVATE(is)
      do is=1,this%n
        this%V(is,j+1) = this%V(is,j+1) - this%V(is,i)*ctmp
      enddo
      this%H(i,j) = this%H(i,j) + ctmp
    enddo
    rtmp = norm(this%V(:,j+1))
    this%H(j+1,j) = rtmp
    rtmp = 1.0_DP/rtmp
!$OMP PARALLEL DO PRIVATE(is)
    do is=1,this%n
      this%V(is,j+1) = this%V(is,j+1)*rtmp
    enddo

    this%j = this%j + 1
    if (this%j > this%m) then
      this%istat = ESLV_OP_END_ARNOLDI
      this%inext = 0
      this%src_vec => NULL()
      this%dst_vec => NULL()
      this%V => NULL()
      this%H => NULL()
      return
    else
      this%istat = ESLV_OP_NOP
      this%inext = 0
      return
    endif
  end select

  return
end subroutine

function norm(x) result(r)
  implicit none
  integer :: i
  complex(DP), intent(in) :: x(:)
  real(DP) :: r
  integer :: n
  n = size(x)
  r = 0.0_DP
!$OMP PARALLEL DO REDUCTION(+:r)
  do i=1,n
    r = r + real(x(i),kind=DP)**2 + aimag(x(i))**2
  enddo
#ifndef _singlePU
  call comlib_sumcast(r)
#endif
  r = sqrt(r)
  return
end function

function prod(x,y) result(c)
  implicit none
  integer :: i
  complex(DP), intent(in) :: x(:),y(:)
  complex(DP) :: c
  integer :: n
  n = size(x)
  c = Z0
!$OMP PARALLEL DO REDUCTION(+:c)
  do i=1,n
    c = c + conjg(x(i)) * y(i)
  enddo
#ifndef _singlePU
  call comlib_sumcast(c)
#endif
  return
end function

subroutine get_schur(icnv,n,A,eval,Z,mode)
  use f95_lapack, only : LA_GEES, LA_HEEV
  use f77_lapack, only : LA_TREXC
  implicit none
  integer, intent(in) :: n,icnv
  complex(DP), intent(inout) :: A(n,n),eval(n),Z(n,n)
  integer, intent(in) :: mode

  complex(DP) :: AA(n-icnv,n-icnv),QQ(n-icnv,n-icnv),ev(n-icnv),ev0(n-icnv)
  real(DP) :: rev(n-icnv)
  integer :: indx(n-icnv),sig((n-icnv)*(n-icnv-1)/2,2)
  integer :: i,j,info,nrem

  nrem = n-icnv
  AA(1:nrem,1:nrem) = A(icnv+1:n,icnv+1:n)
  select case (mode)
  case (ESLV_MODE_GENERAL) 
    !
    ! Q A Q' = AA : uppter triangular
    !           Q : orthgonal basises
    !
    call LA_GEES(AA,ev,VS=QQ)  
  case (ESLV_MODE_HERMITE)
    call LA_HEEV(AA,rev,JOBZ='V',INFO=info)
    QQ(:,:) = AA(:,:)
    ev(:) = rev(:)
  end select
  if (.not.MODE_NORMAL) then
    ev(:) = 1.0_DP/ev(:)
  endif

  ev0(:) = ev(:)
  call bsort(ev,nrem,indx,sig)

  do i=1,nrem
    eval(i+icnv) =  ev0(indx(i))
  enddo

#ifdef _DEBUG
  if (0 == nodeid) then
    do i=1,n
      write(*,'("%TAGS",I3,2E20.9,I3)')i,eval(i)
    enddo
  endif
#endif

  !
  ! sort basis vectors
  !
  select case(mode)
  case (ESLV_MODE_GENERAL)
    do i=1,nrem*(nrem-1)/2
      if (sig(i,1) > 0) then
        call LA_TREXC('V',nrem,AA,nrem,QQ,nrem,sig(i,1),sig(i,2),info)
      else 
        exit
      endif
    enddo
  case (ESLV_MODE_HERMITE)
    Z(:,:)  = Z0
    AA(:,:) = Z0
    do i=1,nrem
      Z(1:nrem,i) = QQ(1:nrem,indx(i))
      AA(i,i) = eval(i+icnv)
    enddo
    QQ(1:nrem,1:nrem) = Z(1:nrem,1:nrem)
  end select

  A(icnv+1:n,icnv+1:n) = AA(1:nrem,1:nrem)
  if (icnv > 0) then
    A(icnv+1:n,1:icnv) = Z0
    A(1:icnv,icnv+1:n) = matmul(A(1:icnv,icnv+1:n),QQ(1:nrem,1:nrem))
  endif

  Z(:,:) = Z0
  Z(icnv+1:n,icnv+1:n) = QQ(1:nrem,1:nrem)
  do i=1,icnv
    Z(i,i) = Z1
  enddo


  return
end subroutine

subroutine sort_eig(icnv,m,eval,n,V)
  implicit none
  integer, intent(in) :: icnv,m,n
  complex(DP), intent(inout) :: eval(m),V(n,m+1)
  integer :: indx(icnv),sig(icnv*(icnv-1)/2,2)
  complex(DP), allocatable :: Q(:,:)
  complex(DP) :: ev(icnv)
  integer :: i
  ev(:) = eval(1:icnv)
  call bsort(ev,icnv,indx,sig)
  allocate(Q(n,icnv))
  ev(:) = eval(1:icnv)
  do i=1,icnv
    eval(i) = ev(indx(i))
    Q(:,i) = V(:,indx(i))
  enddo
  do i=1,icnv
    V(:,i) = Q(:,i)
  enddo
  deallocate(Q)
  return
end subroutine

subroutine get_eig(icnv,n,A,eval,evec,mode)
  use f95_lapack, only : LA_GEEV,LA_HEEV
  implicit none
  integer, intent(in) :: icnv,n
  complex(DP), intent(in)    :: A(n,n)
  complex(DP), intent(inout) :: eval(n)
  complex(DP), intent(inout) :: evec(n,n)
  integer, intent(in) :: mode

  complex(DP) :: AA(n-icnv,n-icnv),hv(n-icnv,n-icnv),ev(n-icnv),ev0(n-icnv)
  real(DP) :: rtmp
  real(DP) :: rev(n-icnv)
  integer :: nrem,i,j,info,indx(n-icnv),sig((n-icnv)*(n-icnv-1)/2,2)

  nrem = n-icnv
  AA(1:nrem,1:nrem) = A(icnv+1:n,icnv+1:n)
  select case (mode)
  case (ESLV_MODE_GENERAL) 
    call LA_GEEV(AA,ev,VR=hv,INFO=info)
  case (ESLV_MODE_HERMITE)
    call LA_HEEV(AA,rev,JOBZ='V',INFO=info)
    hv(:,:) = AA(:,:)
    ev(:) = rev(:)
  end select
  if (.not.MODE_NORMAL) then
    ev(:) = 1.0_DP/ev(:)
  endif

  ev0(:) = ev(:)
  call bsort(ev,nrem,indx,sig)

  do i=1,nrem
    rtmp = sqrt(real(dot_product(hv(:,i),hv(:,i)),kind=DP))
    hv(:,i) = hv(:,i)/rtmp
  enddo

  do i=1,nrem
    eval(i+icnv) = ev0(indx(i))
    evec(1:icnv,i+icnv) = Z0
    evec(icnv+1:n,i+icnv) = hv(1:nrem,indx(i))
  enddo

#ifdef _DEBUG
  if (0 == nodeid) then
    do i=1,n
      write(*,'("%ETAGS",I3,2E20.9,I3)')i,eval(i)
    enddo
  endif
#endif

  return
end subroutine

subroutine get_rank(n,tag,ev)
  implicit none
  integer, intent(in) :: n
  integer, intent(in) :: tag(n)
  complex(DP), intent(inout) :: ev(n)
  integer :: i
  real(DP) :: ev_cnv_max,ev_uncnv_max
  real(DP) :: ev_cnv_min,ev_uncnv_min,rtmp

  ! search max val of conv'd/unconv'd ev
  ev_cnv_max   = 0.0d0
  ev_uncnv_max = 0.0d0
  do i=1,n
    if (tag(i) == EV_CONVED) then
      ev_cnv_max = max(ev_cnv_max,abs(ev(i)))
    endif
    if (tag(i) == EV_UNCONV) then
      ev_uncnv_max = max(ev_uncnv_max,abs(ev(i)))
    endif
  enddo

  ! search min val of unconv'd ev
  ev_cnv_min   = ev_cnv_max
  ev_uncnv_min = ev_uncnv_max
  do i=1,n
    if (tag(i) == EV_CONVED) then
      ev_cnv_min = min(ev_cnv_min,abs(ev(i)))
    endif
    if (tag(i) == EV_UNCONV) then
      ev_uncnv_min = min(ev_uncnv_min,abs(ev(i)))
    endif
  enddo

  if (MODE_NORMAL) then
    if (ev_cnv_min == 0.0_DP) return
    rtmp = 0.9_DP*ev_cnv_min/ev_uncnv_max
  else
    if (ev_cnv_max == 0.0_DP) return
    rtmp = 1.1_DP*ev_cnv_max/ev_uncnv_min
  endif

  do i=1,n
    if (tag(i) == EV_UNCONV) then
      ev(i) = ev(i) * rtmp
    endif
  enddo

  return
end subroutine

subroutine bsort(data,n,indx,sig)
!
! bubble sort
!
  implicit none
  integer,     intent(in) :: n
  complex(DP), intent(inout) :: data(n)
  integer,     intent(inout) :: indx(n),sig(n*(n-1)/2,2)
  complex(DP) :: ctmp
  integer :: i,j,k,ic
  
  sig(:,:)=0
  do i=1,n
    indx(i) = i
  enddo


  ic = 0
  do i=n,1,-1
    do j=1,i-1
      if (MODE_NORMAL) then
        if ( abs(data(j)) < abs(data(j+1)) ) then
          ctmp = data(j)
          data(j)   = data(j+1)
          data(j+1) = ctmp
          k = indx(j)
          indx(j)   = indx(j+1)
          indx(j+1) = k
          ic = ic + 1
          sig(ic,1) = j
          sig(ic,2) = j+1
        endif
      else
        if ( abs(data(j)) > abs(data(j+1)) ) then
          ctmp = data(j)
          data(j)   = data(j+1)
          data(j+1) = ctmp
          k = indx(j)
          indx(j)   = indx(j+1)
          indx(j+1) = k
          ic = ic + 1
          sig(ic,1) = j
          sig(ic,2) = j+1
        endif
      endif
    enddo
  enddo

  return
end subroutine

end module
